@@ -55,7 +55,7 @@ fn output_result_as_tsv(model: &VariantModel, result: &VariantResult, output_lex
     }
 }
 
-fn output_matches_as_json(model: &VariantModel, input: &str, variants: Option<&Vec<VariantResult>>, selected: Option<usize>, offset: Option<Offset>, output_lexmatch: bool, freq_weight: f32, seqnr: usize) {
+fn output_matches_as_json(model: &VariantModel, input: &str, variants: Option<&Vec<VariantResult>>, selected: Option<usize>, offset: Option<Offset>, output_lexmatch: bool, freq_weight: f32, seqnr: usize, tag: Option<u16>, tag_seqnr: Option<u8>) {
     if seqnr > 1 {
         print!("    ,")
     } else {
@@ -65,6 +65,12 @@ fn output_matches_as_json(model: &VariantModel, input: &str, variants: Option<&V
     if let Some(offset) = offset {
         print!(", \"begin\": {}, \"end\": {}", offset.begin, offset.end);
     }
+    if let Some(tag) = tag {
+        print!(", \"tag\": \"{}\"", model.tags.get(tag as usize).expect("tag must exist in model") );
+        if let Some(tag_seqnr) = tag_seqnr {
+            print!(", \"seqnr\": {}",tag_seqnr);
+        }
+    }
     if let Some(variants) = variants {
         println!(", \"variants\": [ ");
         let mut wroteoutput = false;
@@ -237,7 +243,7 @@ fn process(model: &VariantModel, inputstream: impl Read, searchparams: &SearchPa
             }
             let variants = model.find_variants(&input, searchparams);
             if json {
-                output_matches_as_json(model, &input, Some(&variants), Some(0), None, output_lexmatch, searchparams.freq_weight, seqnr);
+                output_matches_as_json(model, &input, Some(&variants), Some(0), None, output_lexmatch, searchparams.freq_weight, seqnr, None, None);
             } else {
                 //Normal output mode
                 output_matches_as_tsv(model, &input, Some(&variants), Some(0), None,  output_lexmatch, searchparams.freq_weight);
@@ -276,7 +282,7 @@ fn process_par(model: &VariantModel, inputstream: impl Read, searchparams: &Sear
         for (input, variants) in output {
             seqnr += 1;
             if json {
-                output_matches_as_json(model, &input, Some(&variants), Some(0), None, output_lexmatch, searchparams.freq_weight, seqnr);
+                output_matches_as_json(model, &input, Some(&variants), Some(0), None, output_lexmatch, searchparams.freq_weight, seqnr, None, None);
             } else {
                 //Normal output mode
                 output_matches_as_tsv(model, &input, Some(&variants), Some(0), None, output_lexmatch, searchparams.freq_weight);
@@ -395,7 +401,7 @@ fn process_search(model: &VariantModel, inputstream: impl Read, searchparams: &S
         for result_match in output {
             seqnr += 1;
             if json {
-                output_matches_as_json(model, result_match.text, result_match.variants.as_ref(), result_match.selected, Some(result_match.offset), output_lexmatch, searchparams.freq_weight, seqnr);
+                output_matches_as_json(model, result_match.text, result_match.variants.as_ref(), result_match.selected, Some(result_match.offset), output_lexmatch, searchparams.freq_weight, seqnr, result_match.tag, result_match.seqnr);
             } else {
                 //Normal output mode
                 output_matches_as_tsv(model, result_match.text, result_match.variants.as_ref(), result_match.selected, Some(result_match.offset), output_lexmatch, searchparams.freq_weight);
@@ -557,6 +563,10 @@ pub fn common_arguments<'a,'b>() -> Vec<clap::Arg<'a,'b>> {
         .help("Number of matches to return per input (set to 0 for unlimited if you want to exhaustively return every possibility within the specified anagram and edit distance)")
         .takes_value(true)
         .default_value("10"));
+    args.push(Arg::with_name("unicodeoffsets")
+        .long("unicode-offsets")
+        .short("u")
+        .help("Output all text offsets in unicode points rather than UTF-8 byte offsets"));
     args.push(Arg::with_name("files")
         .help("Input files")
         .takes_value(true)
@@ -606,6 +616,11 @@ pub fn search_arguments<'a,'b>() -> Vec<clap::Arg<'a,'b>> {
             .long("weight-variant-model")
             .help("Weight attributed to the variant model in finding the most likely sequence in search mode")
             .takes_value(true)
+            .default_value("3.0"));
+        args.push(Arg::with_name("weight-contextrules")
+            .long("weight-contextrules")
+            .help("Weight attributed to the context rules model in finding the most likely sequence in search mode")
+            .takes_value(true)
             .default_value("1.0"));
         args.push(Arg::with_name("weight-context")
             .long("weight-context")
@@ -809,11 +824,17 @@ fn main() {
         } else {
             1.0
         },
+        contextrules_weight: if args.is_present("weight-contextrules") {
+            args.value_of("weight-contextrules").unwrap().parse::<f32>().expect("Context-rules weight should be a floating point number")
+        } else {
+            1.0
+        },
         max_seq: if args.is_present("max-seq") {
             args.value_of("max-seq").unwrap().parse::<usize>().expect("max-seq must be an integer")
         } else {
             250
         },
+        unicodeoffsets: args.is_present("unicode-offsets")
     };
 
 
